#!/usr/bin/env python3

"""Chained CI config wrapper."""
import copy
import os
import pkg_resources
from yaml import load, SafeLoader, add_constructor
import ansible.parsing.yaml.objects
from ansible.module_utils._text import to_bytes
from ansible.parsing.dataloader import DataLoader
from ansible.parsing.vault import VaultSecret
from chainedci.env_vars import EnvVars


# pylint: disable=invalid-name
__configDict = {}


def getConfig(name, load_env=True, init_scenarios=False, reset=False):
    """
    Get the config object with its name.

    :param name: The name of the config to get
    :type name: string
    :param load_env: Load or not environment variables if config does not
                     exists
    :type load_env: Boolean
    :param init_scenarios: Load or not scenarios if config does not exists
    :type init_scenarios: Boolean
    :return: a Config Class
    :rtype: Config object
    """
    if name in __configDict and not reset:
        return __configDict[name]
    __configDict[name] = Config(load_env, init_scenarios)
    return __configDict[name]


def merge_dicts(default, user):
    """Overload default dictionnary with the user one."""
    default = copy.deepcopy(default)
    user = copy.deepcopy(user)
    if isinstance(user, dict) and isinstance(default, dict):
        for k, v in default.items():
            if k not in user:
                user[k] = v
            else:
                user[k] = merge_dicts(v, user[k])
    return user


def init_ini(default_config_file='default_chainedci.yml',
             user_config_file='chainedci.yml'):
    """Initialize the ini config.

    with the defaults overloaded with the user config.
    """
    ini = getConfig('ini', reset=True)
    default_config = pkg_resources.resource_string('chainedci',
                                                   default_config_file)
    ini.load_content(default_config)
    if os.path.exists(user_config_file):
        ini.merge(user_config_file)


def simple_dict(data):
    """Recursive function to parse AnsibleVaultEncryptedUnicode as str."""
    if isinstance(data, dict):
        for k, v in data.items():
            data[k] = simple_dict(v)
    elif isinstance(data, list):
        # pylint: disable=consider-using-enumerate
        for i in range(0, len(data)):
            data[i] = simple_dict(data[i])
    elif isinstance(data,
                    ansible.parsing.yaml.objects.AnsibleVaultEncryptedUnicode):
        data = str(data)
    return data


def _create_vault_construct(loader, node):
    """Yaml constructor for vault."""
    loader.construct_scalar(node)
    return "__VAULTED_VALUE__"


def _load_yaml_data(data):
    """Load data without ansible loader. Assiming it is YAML."""
    add_constructor(
        '!vault', _create_vault_construct, SafeLoader)
    return load(data, SafeLoader)


class Config(dict):
    """Config class compatible with ansible vaulted files."""

    dl = DataLoader()

    def __init__(self, load_env=True, init_scenarios=False):
        """
        Init config class.

        Add a env attribute to access Env Vars
        """
        super().__init__()
        if load_env:
            self['env'] = EnvVars()
        if init_scenarios:
            self['defaults'] = None
            self['scenarios'] = {}

    def load_vault_key(self, filename):
        """
        Set vault secret from file.

        :param filename: the file name to load with path
        :type filename: string

        :Example:
        >>> c = Config()
        >>> c.load_vault_key('tests/config/config_example_vault_file')
        """
        with open(filename, encoding="utf-8") as f:
            content = f.readlines()
            self.set_vault_key(content[0].rstrip())

    def get_vault_key(self):
        """
        Get vault secret.

        :return: the vault key
        :rtype: VaultLib

        :Example:
        >>> c = Config()
        >>> c.load_vault_key('tests/config/config_example_vault_file')
        >>> type(c.get_vault_key())
        <class 'ansible.parsing.vault.VaultLib'>
        """
        # pylint: disable=protected-access
        return self.__class__.dl._vault

    def set_vault_key(self, key, key_id='default'):
        """
        Set vault secret.

        :param key: the vault key
        :type key: string

        :param key_id: the vault key name
        :type key_id: string

        :Example:
        >>> c = Config()
        >>> c.set_vault_key('fake')
        """
        secret = [(key_id, VaultSecret(to_bytes(key, encoding='utf-8')))]
        self.__class__.dl.set_vault_secrets(secret)

    def clean_vault_keys(self):
        """Remove all vault keys."""
        # pylint: disable=protected-access
        type(self).dl = DataLoader()

    def load_file(self, filename, key_name=None):
        """
        Load config file.

        :param filename: the file name to load with path
        :type filename: string

        :Example:
        >>> c = Config()
        >>> c.load_file('tests/config/config_example.yml')
        """
        if self.get_vault_key().secrets:
            content = self.__class__.dl.load_from_file(filename)
        else:
            with open(filename, encoding="utf-8") as f:
                content = _load_yaml_data(f)
        self.update_key(content, key_name)

    def load_content(self, content, key_name=None):
        """
        Load config file.

        :param content: the config content to load
        :type content: string
        """
        if self.get_vault_key().secrets:
            content = self.__class__.dl.load(content)
        else:
            content = _load_yaml_data(content)
        self.update_key(content, key_name)

    def update_key(self, content, key_name=None):
        """
        Update content with an optional key_name.

        :param content: the config content to load
        :type content: string
        :param key_name: the optionan key to update, if none, update self
        :type content: string
        """
        if key_name is not None:
            dic = self
            keys = key_name.split('.')
            for key in keys[:-1]:
                dic = dic.setdefault(key, {})
            dic[keys[-1]] = content
            self.update(dic)
        else:
            self.update(content)

    def merge(self, newfile, key_name=None):
        """Merge a config from a file to this object."""
        new = self.__class__.dl.load_from_file(newfile) or {}
        actual = self[key_name] if key_name else self
        self.update_key(merge_dicts(actual, new), key_name)

    def simplify(self):
        """Remove ansible objects in this object."""
        self.update(simple_dict(self))


# Load ini file
# pylint: disable=invalid-name
config = getConfig('chainedci', load_env=False)
run = getConfig('run', load_env=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2019 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Chained CI environment variables wrapper."""

import os
import re
import logging


class EnvVars(dict):
    """Class to access environment variables."""

    def __init__(self, load=True):
        """
        Show Environmnent variable as a dict with object access.

        :param load: Load or not env vars at init
        :type load: boolean
        :return: a EnvVars object
        :rtype: EnvVars class

        >>> env = EnvVars()
        >>> env['PATH']
        '.../usr/bin:...'
        >>> env.PATH
        '.../usr/bin:...'
        >>> env
        {...'PATH': .../usr/bin:...}
        >>> 'HOME' in env
        True
        >>> 'Zorglub' in env
        False
        >>> env.PATH
        '.../usr/bin:...'
        """
        if load:
            super().__init__(dict(os.environ))
        else:
            super()

    def eval(self, condition):
        """
        Eval a condition string on the environment vars.

        :param condition: The condition to eval
            The condition can be:
            - a variable presence check: 'VAR_NAME'
            - an equality: 'VAR_NAME == VALUE'
            - a difference: 'VAR_NAME != VALUE'
            - a test the value is in a list 'VAR_NAME in [VAL1, VAL2]'
            In any case, if the variable does not exists, the result is false.
        :type condition: string
        :return: the eval result
        :rtype: boolean

        >>> env = EnvVars()
        >>> env['USER']
        johndoe
        >>> env.eval('USER')
        True
        >>> env.eval('USER == johndoe')
        True
        >>> env.eval('USER == janedoe')
        False
        >>> env.eval('USER != janedoe')
        True
        >>> env.eval('USER in [johndoe, janedoe]')
        True
        >>> env.eval('USER in [janedoe, bobdoe]')
        False
        """
        log = logging.getLogger('chainedci')
        if all(sep not in condition for sep in [' ', '=']):
            ret = condition in self
            log.debug('env_condition_parser - check if %s in env -> %s',
                      condition, ret)
        elif '=' in condition:
            (var, equ, val) = re.findall(r'^(.*[^\s])\s*([=\!]=)\s*([^\s].*)$',
                                         condition)[0]
            if var not in self:
                log.debug('env_condition_parser - %s not in env -> %s',
                          var, False)
                return False
            ret = (self[var] == val) if (equ == '==') else (self[var] != val)
            log.debug('env_condition_parser - eval equality '
                      '"%s" %s "%s" -> %s', var, equ, val, ret)
        elif ' in ' in condition:
            (var, val) = re.findall(r'^(.*[^\s])\s* in '
                                    r'\s*\[\s*([^\s].*[^\s])\s*\]$',
                                    condition)[0]
            val = re.sub(r'\s*,\s*', ',', val).split(',')
            if var not in self:
                log.debug('env_condition_parser - %s not in env -> %s',
                          var, False)
                return False
            ret = self[var] in val
            log.debug('env_condition_parser - eval '
                      '"%s" in %s -> %s', var, val, ret)
        else:
            log.warning('condition not parsed: %s', condition)
            ret = False
        return ret



#!/usr/bin/env python3

"""Logging functions and vars."""

import logging
import sys
from chainedci.config import getConfig
from chainedci.parser import JinjaParser
from chainedci.tools import raise_ex


def init_log(ini):
    """Config the default logging."""
    console_handler = logging.StreamHandler(sys.stdout)
    log_handlers = [console_handler]
    if ini['log']['file']:
        file_handler = logging.FileHandler(ini['log']['file']['name'])
        log_handlers.append(file_handler)
    logging.basicConfig(format=ini['log']['format'],
                        handlers=log_handlers,
                        level=ini['log']['level'])


def deprecation_warning(msg):
    """Print a deprecation message."""
    ini = getConfig('ini')
    if ini['log']['deprecation_warning']:
        log.warning("DEPRECATION: %s", msg)
        print(f"DEPRECATION: {msg}")


class TokenFilter(logging.Filter):
    """Logging filter to avoid token leaking."""

    def __init__(self, patterns):
        """
        Init with a list of patterns to hide.

        :param patterns: Patterns to hide
        :type patterns: list
        """
        super().__init__()
        self._patterns = patterns

    def filter(self, record):
        """Filter the message."""
        record.msg = self.redact(record.msg)
        if isinstance(record.args, dict):
            for k in record.args.keys():
                record.args[k] = self.redact(record.args[k])
        else:
            record.args = tuple(self.redact(arg) for arg in record.args)
        return True

    def redact(self, msg):
        """Replace patterns with a message."""
        msg = msg if isinstance(msg, str) else str(msg)
        for pattern in self._patterns:
            if isinstance(pattern, (bytes, bytearray)):
                pattern = pattern.decode("utf-8")
            msg = msg.replace(pattern, "__HIDDEN__")
        return msg


def add_tokens_to_log_filter(conf, additional_tokens=None):
    """
    Add all tokens to the log filter.

    This will fetch all tokens in all.yml and add them to log filter.
    Token from env vars can be added by setting the list of the vars setting
    the param additional_env_tokens_names.

    If the values of one of the tokens is empty it will raise an exception.

    :param conf: a config object
    :type conf: Config object
    :param additional_tokens: the list of tokens to hide
    :type additional_tokens: list of string
    """
    ini = getConfig('ini')
    if not ini['log']['hide_tokens']:
        log.info("NO filter added to the log system due to log/hide_tokens "
                 "bypass")
        return
    log.info("Add filters to the log system")
    token_list = []
    # Add global tokens to filters
    if 'defaults' in conf:
        if 'tokens' in conf['defaults']:
            for token in conf['defaults']['tokens']:
                log.debug("add_tokens_to_log_filter - Add %s to token log "
                          "filter", token['name'])
                parser = JinjaParser()
                token_value = parser.parse(token['value'], True)
                if token_value == '':  # nosec B105
                    raise_ex(ValueError, f"Please set a value for token "
                                         f"{token['name']} in config file "
                                         "all.yml")
                log.debug("add_tokens_to_log_filter - type of token value: %s",
                          type(token_value))
                token_list.append(token_value)
        # Add project tokens to filters
        if 'gitlab' in conf['defaults']:
            for project_name in conf['defaults']['gitlab']['git_projects']:
                project = conf['defaults']['gitlab']['git_projects'][
                    project_name]
                if 'trigger_token' in project:
                    parser = JinjaParser()
                    token_value = parser.parse(str(project['trigger_token']),
                                               True)
                    if token_value == '':  # nosec B105
                        raise_ex(ValueError,
                                 "Please set a value for 'trigger_token' "
                                 f"of project {project_name} in config "
                                 "file all.yml")
                    log.debug("add_tokens_to_log_filter - Add project '%s' "
                              "trigger_token to token log filter",
                              project_name)
                    token_list.append(token_value)
    # Add additional tokens
    for token in additional_tokens or []:
        log.debug("add_tokens_to_log_filter - "
                  "Add additional token to log filter")
        token_list.append(token)
    log.addFilter(TokenFilter(token_list))


# pylint: disable=invalid-name
log = logging.getLogger('chainedci')
